from __future__ import absolute_import

import os
import numpy as np
import tensorflow_hub as hu
import sklearn.preprocessing as skp
import shelve as she

from ..utilities import imagehelpers as ih
from ..utilities import httphelpers as hh
from ..utilities import filehelpers as fh

NNLM_EN_DIM128 = {
    'input_size': None,
    'input_type': 'text',
    'output_size': 128,
    'model_url': 'https://tfhub.dev/google/nnlm-en-dim128/2'
}

UNIVERSAL_SENTENCE_ENCODER_LARGE = {
    'input_size': None,
    'input_type': 'longtext',
    'output_size': 512,
    'model_url': 'https://tfhub.dev/google/universal-sentence-encoder-large/5'
}

IMAGENET_INCEPTION_V3_FEATURE_VECTOR = {
    'input_size': (299, 299),
    'input_type': 'image_url',
    'output_size': 2048,
    'model_url': 'https://tfhub.dev/google/imagenet/inception_v3/feature_vector/4'
}


class BaseEmbedder:

    def __init__(self, input_type=None, input_size=None, output_size=None, cache_path=None):

        self.input_type = input_type
        self.input_size = input_size
        self.output_size = output_size
        self.cache_path = cache_path
        self._shelve = None

        self.load_cache()

    def __exit__(self, *exc_info):
        self.save_cache()

    def load_cache(self):
        
        cache_path = None

        if os.path.exists(self.cache_path):
            cache_path = os.path.join(
                self.cache_path, 
                self.input_type)
        else:
            cache_path = self.input_type

        self._shelve = she.open(
            filename=cache_path)

    def save_cache(self):
        self._shelve.close()

    def flush_cache(self):
        self.save_cache()
        self.load_cache()

    def data_empty(self, data, feature):
        return data

    def data_preparation(self, data, feature):
        return data

    def data_embedding(self, data, feature):
        return data

    def embed_data(self, data, feature):
        if data is None:
            return self.data_empty(
                data=data, 
                feature=feature)
        
        data_prepared = self.data_preparation(
            data=data, 
            feature=feature)

        embedding = None

        if data in self._shelve:
            embedding = self._shelve[data]
        else:
            embedding = self.data_embedding(
                data=data_prepared, 
                feature=feature)
            
            self._shelve[data] = embedding
        
        return embedding


class CategoryEmbedder(BaseEmbedder):

    def __init__(self, input_size=None, output_size=None, cache_path=None):
        super().__init__(
            input_type='category',
            input_size=input_size,
            output_size=output_size,
            cache_path=cache_path)

    def data_empty(self, data, feature):
        empty_data = None

        if self.output_size is None:
            empty_data = [0] * len(feature.classes)
        else:
            empty_data = [0] * self.output_size

        return np.array(
            empty_data,
            dtype=np.float64)

    def data_embedding(self, data, feature):
        vect = self.data_empty(data, feature)

        if data is not None:
            position = feature.classes.index(data)

            if position < len(vect):
                vect[position] = 1

        return vect


class NumericEmbedder(BaseEmbedder):

    def __init__(self, default_value=0.0, cache_path=None):
        super().__init__(
            input_type='numeric',
            input_size=None,
            output_size=None, 
            cache_path=cache_path)

        self.default_value = default_value

    def data_empty(self, data, feature):
        return np.array(
            [self.default_value],
            dtype=np.float64)

    def data_embedding(self, data, feature):
        return np.array(
            [data],
            dtype=np.float64)


class BaseHubEmbedder(BaseEmbedder):

    def __init__(self, input_type=None, input_size=None, output_size=None, model_url=None, cache_path=None):
        super().__init__(
            input_type=input_type,
            input_size=input_size,
            output_size=output_size, 
            cache_path=cache_path)

        self.model = hu.load(model_url)
        self._empty = np.array([0]*output_size, np.float64)

    def data_empty(self, data, feature):
        return self._empty

    def data_embedding(self, data, feature):
        return skp.normalize(
            self.model(data), 
            axis=1)[0]


class TextEmbedder(BaseHubEmbedder):

    def data_preparation(self, data, feature):
        input = data
        if self.input_size is not None and len(data) > self.input_size:
            input = data[:self.input_size]
            
        return np.array([input])


class ImageUrlEmbedder(BaseHubEmbedder):

    def data_preparation(self, data, feature):
        pil_image = hh.download_image(
            url=data)
        
        if pil_image.size != self.input_size:
            pil_image = ih.extract_square_portion(
                pil_image=pil_image,
                output_size=self.input_size)

        return ih.image_to_batch_array(
            pil_image=pil_image,
            rescaled=True)



import math as m
import pandas as pd
import numpy as np

import matplotlib as mat
from matplotlib import pyplot as plt
import pylab as pyl
import seaborn as sns

from scipy import optimize as sco

import sklearn as sk
from sklearn import exceptions as skx
from sklearn import preprocessing as skp
from sklearn import model_selection as skms
from sklearn import pipeline as skpl
from sklearn import decomposition as skd
from sklearn import linear_model as sklm
from sklearn import ensemble as skle
from sklearn import neighbors as skln
from sklearn import dummy as sky
from sklearn import metrics as skm
from sklearn import calibration as skc
from sklearn.utils import validation as skuv
from sklearn.utils import class_weight as skcw

from . import commonhelpers as cmn
from . import graphichelpers as gph



class MulticlassClassifierOptimizer(object):

    def __init__(self, model, classes, scoring_function):

        if not MulticlassClassifierOptimizer.optimizable_model(model):
            raise NotImplementedError('The model does not implement method "predict_proba" and cannot be optimized.')

        self.model = model
        self.classes = classes
        self.scoring_function = scoring_function

        self.optimized = False
        self.thresholds = None

    @property
    def estimator_(self):

        self.assert_optimized()

        return MulticlassClassifierOptimizer.get_estimator(
            model=self.model)

    @property
    def classes_(self):
        return self.estimator_.classes_

    @property
    def coef_(self):
        return self.estimator_.coef_

    @property
    def intercept_(self):
        return self.estimator_.intercept_

    def fit(self, X, y):

        if not MulticlassClassifierOptimizer.fitted_model(self.model):

            print('      -> Fitting base model (wasn\'t fitted).')

            self.model.fit(
                X=X,
                y=y)

        print('      -> Model calibration.')

        self.model = skc.CalibratedClassifierCV(
            base_estimator=self.model,
            method='sigmoid',
            cv='prefit')

        self.model.fit(
            X=X,
            y=y,
            sample_weight=skcw.compute_sample_weight(
                class_weight='balanced', 
                y=y))

        print('      -> Optimizing multiclass thresholds.')

        self.thresholds = MulticlassClassifierOptimizer.get_optimized_thresholds(
            scoring_function=self.scoring_function,
            y_true=MulticlassClassifierOptimizer.one_hot_encode(
                y=y),
            y_score=self.model.predict_proba(
                X=X))

        self.optimized = True

        return self

    def predict_proba(self, X):
        """
        Predict class probabilities for X.

        Args:
            X (ndarray, shape (n_samples, n_features)): The input samples.
        
        Returns:
            ndarray, shape (n_samples, n_classes): The class probabilities 
            of the input samples. The order of the classes corresponds to that
            in the attribute :term:`classes_`.
        """
        self.assert_optimized()

        return self.model.predict_proba(X)

    def predict(self, X):
        """Predict class labels for samples in X, by using the
        `predict_proba` model function and the optimized `thresholds`
        values (class specific)

        Args:
            X (ndarray, shape (n_samples, n_features)): Input data for prediction.

        Returns:
            ndarray, shape (n_samples, n_classes): Predicted class label per sample (1-hot encoded).
        """
        self.assert_optimized()

        return self.predict_from_score(
            thresholds=self.thresholds,
            y_score=self.model.predict_proba(X))

    def score(self, X, y):

        self.assert_optimized()

        class_weights = 1/skcw.compute_class_weight(
            class_weight='balanced', 
            classes=np.arange(
                start=0,
                stop=len(self.classes),
                step=1,
                dtype=int), 
            y=y)

        y_true = MulticlassClassifierOptimizer.one_hot_encode(y=y)
        y_score = self.predict_proba(X)

        scores = []

        for index, _ in enumerate(self.classes):

            scores.append(
                self.scoring_function(
                    threshold=self.thresholds[index],
                    y_true=y_true[:, index],
                    y_score=y_score[:, index]))

        return np.average(
            a=scores,
            weights=class_weights)

    def assert_optimized(self):
        if not self.optimized:
            raise NotImplementedError('The model has not been optimized yet.')

    def predict_from_score_1c(self, y_score, class_index, class_threshold):
        updated_thresholds = self.thresholds.copy()
        updated_thresholds[class_index] = class_threshold

        y_pred = MulticlassClassifierOptimizer.predict_from_score(
            thresholds=updated_thresholds,
            y_score=y_score)

        return y_pred[:, class_index]

    def confusion_matrix_1c(self, y_true, y_score, class_index, class_threshold):
        return skm.confusion_matrix(
            y_true=y_true[:, class_index], 
            y_pred=self.predict_from_score_1c(
                y_score=y_score, 
                class_index=class_index, 
                class_threshold=class_threshold),
            labels=[0.0, 1.0]).ravel()

    def sort_classes_by_score(self, X, y, n_top=10):

        y_true = MulticlassClassifierOptimizer.one_hot_encode(y)
        y_score = self.predict_proba(X)

        records = []

        for index, name in enumerate(self.classes):
            records.append({
                'Class': name,
                'Index': index,
                'Score': self.scoring_function(
                    self.thresholds[index],
                    y_true[:, index], 
                    y_score[:, index])
            })

        df_scores = pd.DataFrame.from_records(records)
        df_scores = df_scores.convert_dtypes()
        df_scores = df_scores.sort_values(by='Score', ascending=True)

        return list(df_scores.head(n_top)['Class'].values)

    def get_metrics_by_class(self, X, y, transpose=False):
        return self.get_metrics_by_class_base(
            y_true=MulticlassClassifierOptimizer.one_hot_encode(y),
            y_score=self.predict_proba(X),
            transpose=transpose)

    def get_metrics_by_class_base(self, y_true, y_score, transpose=False):
        
        records = []

        for index, name in enumerate(self.classes):

            tn, fp, fn, tp = self.confusion_matrix_1c(
                y_true=y_true,
                y_score=y_score,
                class_index=index,
                class_threshold=self.thresholds[index])

            record = {
                'Class': name,
                'True Negative': tn,
                'False Positive': fp,
                'False Negative': fn,
                'True Positive': tp,
                'Prevalence': BinaryClassifierHelper.prevalence(tn, fp, fn, tp),
                'Actual Negative': tn + fp,
                'Actual Positive': fn + tp,
                'Predicted Negative': tn + fn,
                'Predicted Positive': fp + tp,
                'Predicted Correctly': tn + tp,
                'Predicted Incorrectly': fn + fp,
                'Accuracy': float(tn + tp) / float(tn + fp + fn + tp),
                'Precision': BinaryClassifierHelper.precision(tn, fp, fn, tp),
                'Recall': BinaryClassifierHelper.recall(tn, fp, fn, tp),
                'F1-Score': BinaryClassifierHelper.f1(tn, fp, fn, tp),
                'Fallout': BinaryClassifierHelper.fallout(tn, fp, fn, tp),
                'ROC AUC Score': skm.roc_auc_score(y_true[:, index], y_score[:, index]),
                'Brier Score': skm.brier_score_loss(y_true[:, index], y_score[:, index]),
                'Cross-Entropy': skm.log_loss(y_true[:, index], y_score[:, index]),
            }

            records.append(record)

        df_scores = pd.DataFrame.from_records(records)
        df_scores = df_scores.set_index('Class')
        df_scores = df_scores.convert_dtypes()

        if transpose:
            return df_scores.transpose()
        else:
            return df_scores

    def plot_curve(self, class_name, X, y, n_bins=10):

        if class_name in self.classes:

            y_true = MulticlassClassifierOptimizer.one_hot_encode(y)
            y_score = self.predict_proba(X)

            index = self.classes.index(class_name)

            BinaryClassifierHelper.plot_curves(
                title='{0}\nScore: {1:.4f}'.format(
                    class_name.upper(), 
                    self.scoring_function(
                        threshold=self.thresholds[index],
                        y_true=y_true[:, index],
                        y_score=y_score[:, index])),
                y_true=y_true[:, index], 
                y_score=y_score[:, index],
                best_threshold=self.thresholds[index],
                n_bins=n_bins)

    def plot_curves(self, X, y, n_bins=10, n_top=10):

        classes = None

        if n_top is not None:
            classes = self.sort_classes_by_score(
                X=X,
                y=y,
                n_top=n_top)

        self.plot_curves_base(
            y_true=MulticlassClassifierOptimizer.one_hot_encode(y),
            y_score=self.predict_proba(X),
            n_bins=n_bins,
            classes=classes)

    def plot_curves_base(self, y_true, y_score, n_bins=10, classes=None):
        
        for index, name in enumerate(self.classes):
            if (classes is None or len(classes) == 0) or (classes is not None and name in classes):
                BinaryClassifierHelper.plot_curves(
                    title='{0}\nScore: {1:.4f}'.format(
                        name.upper(), 
                        self.scoring_function(
                            threshold=self.thresholds[index],
                            y_true=y_true[:, index],
                            y_score=y_score[:, index])),
                    y_true=y_true[:, index], 
                    y_score=y_score[:, index],
                    best_threshold=self.thresholds[index],
                    n_bins=n_bins)

    @staticmethod
    def optimizable_model(model):
        return callable(getattr(model, 'predict_proba', None))

    @staticmethod
    def predict_from_score(thresholds, y_score):
        """Predict class labels for scores and the optimized `thresholds`
        values (class specific)

        Args:
            y_score (ndarray, shape (n_samples, n_classes)): Input data for prediction.

        Returns:
            ndarray, shape (n_samples, n_classes): Predicted class label per sample (1-hot encoded).
        """
        # Find the winners (scores higher than class-threshold).
        y_pred = np.array(
            (np.array(y_score) > thresholds),
            dtype=float)

        # Number of winners for each sample
        y_n_winners = y_pred.sum(
            axis=1,
            dtype=float)
        
        # Ratio of the scores compared to each class-threshold (how far are we from the threshold)
        y_t_ratio = np.array(
            y_score / thresholds,
            dtype=float)
        
        # Samples with no clear winner: proba ratio considering all classes (none beyond threshold)
        y_0w_vect = (1 - y_pred) * y_score * y_t_ratio

        # Samples with no clear winner: the class with the highest % wins
        y_0w_winner = np.array(
            (y_0w_vect == y_0w_vect.max(axis=1).reshape(-1, 1)),
            dtype=float)
        
        #Samples with multiple winners: proba ratio considering winning classes only (above threshold)
        y_nw_vect = y_pred * y_score * y_t_ratio
        
        # Samples with multiple winners: the class with the highest % wins
        y_nw_winner = np.array(
            (y_nw_vect == y_nw_vect.max(axis=1).reshape(-1, 1)),
            dtype=float)
        
        return np.where(
            (y_n_winners.reshape(-1, 1) == 1),
            y_pred, # The easy ones having 1 true winner
            np.where(
                (y_n_winners.reshape(-1, 1) == 0),
                y_0w_winner, # There is no winner
                y_nw_winner)) # There are multiple winners

    @staticmethod
    def one_hot_encode(y):
        y_true = np.zeros(
            shape=(y.shape[0], y.max() + 1),
            dtype=float)

        for i in range(y.shape[0]):
            y_true[i, y[i]] = 1.0

        return y_true

    @staticmethod
    def get_optimized_thresholds(scoring_function, y_true, y_score):
        
        # The objective function is the 
        objective = lambda threshold, y_true, y_score: 1 - scoring_function(threshold, y_true, y_score)

        # Define alpha
        alpha = 0.05

        # Get the threshold value minimizing the score function for each class
        thresholds = []
        
        for i in range(y_true.shape[1]):
            
            result = sco.minimize_scalar(
                fun=objective, 
                bounds=(0 + alpha, 1 - alpha), 
                args=(y_true[:, i], y_score[:, i]),
                method='bounded',
                options={
                    'xatol': 1e-5, 
                    'maxiter': 250 })

            thresholds.append(
                result.x)
                
        return np.array(thresholds)

    @staticmethod
    def get_estimator(model):
        
        estimator = model

        if type(model).__name__ == 'Pipeline':
            _, estimator = model.steps[-1]

        return estimator

    @staticmethod
    def fitted_model(model):
        
        estimator = MulticlassClassifierOptimizer.get_estimator(
            model=model)
        
        try:
            skuv.check_is_fitted(estimator)
            return True
        except skx.NotFittedError:
            return False


class BinaryClassifierHelper(object):

    @staticmethod
    def confusion_matrix(threshold, y_true, y_score):
        return skm.confusion_matrix(
            y_true=y_true, 
            y_pred=np.array(
                (np.array(y_score) > threshold),
                dtype=float),
            labels=[0.0, 1.0]).ravel()

    @staticmethod
    def f1_score(threshold, y_true, y_score):
        return BinaryClassifierHelper.f1(
            *BinaryClassifierHelper.confusion_matrix(
                threshold=threshold,
                y_true=y_true, 
                y_score=y_score))

    @staticmethod
    def get_confusion_matrix_string(threshold, y_true, y_score):
        tn, fp, fn, tp = BinaryClassifierHelper.confusion_matrix(
            threshold=threshold,
            y_true=y_true,
            y_score=y_score)

        metrics = [
            ('Threshold', threshold, '.4f'),
            ('TN', tn, 'd'),
            ('FP', fp, 'd'),
            ('FN', fn, 'd'),
            ('TP', tp, 'd')
        ]
            
        return cmn.flatten_namevalue_pairs(
            pairs=metrics,
            separator='\n')

    @staticmethod
    def get_metrics_string(threshold, y_true, y_score):
        tn, fp, fn, tp = BinaryClassifierHelper.confusion_matrix(
            threshold=threshold,
            y_true=y_true,
            y_score=y_score)
        
        metrics = [
            ('Accuracy', BinaryClassifierHelper.accuracy(tn, fp, fn, tp), '.4f'),
            ('Recall', BinaryClassifierHelper.recall(tn, fp, fn, tp), '.4f'),
            ('Precision', BinaryClassifierHelper.precision(tn, fp, fn, tp), '.4f'),
            ('Fallout', BinaryClassifierHelper.fallout(tn, fp, fn, tp), '.4f'),
            ('Prevalence', BinaryClassifierHelper.prevalence(tn, fp, fn, tp), '.4f'),
            ('ROC AUC Score', skm.roc_auc_score(y_true, y_score), '.4f'),
            ('Brier Score', skm.brier_score_loss(y_true, y_score), '.4f'),
            ('Cross-Entropy', skm.log_loss(y_true, y_score), '.4f')
        ]
            
        return cmn.flatten_namevalue_pairs(
            pairs=metrics,
            separator='\n')

    @staticmethod
    def recall(tn, fp, fn, tp):
        if tp + fn == 0:
            return float(1)
        else:
            return float(tp) / float(tp + fn)

    @staticmethod
    def fallout(tn, fp, fn, tp):
        if tn == 0:
            if fp > 0:
                return float(1)
            else:
                return float(0)
        else:
            return float(fp) / float(fp + tn)

    @staticmethod
    def precision(tn, fp, fn, tp):
        if tp == 0:
            if fp > 0:
                return float(0)
            else:
                return float(1)
        else:
            return float(tp) / float(tp + fp)

    @staticmethod
    def f1(tn, fp, fn, tp):
        den = float(tp) + .5*(float(fp + fn))

        if den == 0.0:
            return float(1)
        else:
            return float(tp) / den

    @staticmethod
    def accuracy(tn, fp, fn, tp):
        if tn + fp + fn + tp == 0:
            return float(0)
        else:
            return float(tp + tn) / float(tn + fp + fn + tp)

    @staticmethod
    def prevalence(tn, fp, fn, tp):
        if tn + fp + fn + tp == 0:
            return float(0)
        else:
            return float(tp + fn) / float(tn + fp + fn + tp)

    @staticmethod
    def calculate_tpr_fpr_prec(threshold, y_true, y_score):
        tn, fp, fn, tp = BinaryClassifierHelper.confusion_matrix(
            threshold=threshold,
            y_true=y_true,
            y_score=y_score)
        
        recall_ = BinaryClassifierHelper.recall(tn, fp, fn, tp)
        fallout_ = BinaryClassifierHelper.fallout(tn, fp, fn, tp)
        precision_ = BinaryClassifierHelper.precision(tn, fp, fn, tp)
        
        return recall_, fallout_, precision_

    @staticmethod
    def plot_curves(title, y_true, y_score, best_threshold=None, n_bins=10):

        # syle
        gph.GraphicsStatics.initialize_matplotlib_styles()

        roc_color = 'crimson'
        pr_color = 'royalblue'
        main_linestyle = 'solid'
        neutral_color = 'k'
        neutral_linestyle = 'dashed'
        lw = 2

        fig = plt.figure(
            figsize=gph.GraphicsStatics.g_portrait_fig_size)

        ax = []
        
        gs = fig.add_gridspec(3, 2)
        
        ax.append(
            fig.add_subplot(gs[0, 0]))
        
        ax.append(
            fig.add_subplot(gs[0, 1]))
        
        ax.append(
            fig.add_subplot(gs[1, :]))
        
        ax.append(
            fig.add_subplot(gs[2, 0]))
        
        ax.append(
            fig.add_subplot(gs[2, 1]))

        fpr, tpr_recall, _ = skm.roc_curve(
            y_true, 
            y_score,
            pos_label=1)
        
        ax[0].step(
            fpr,
            tpr_recall,
            color=roc_color,
            alpha=0.2,
            where='post',
            linewidth=lw,
            linestyle=main_linestyle)
        
        ax[0].fill_between(
            fpr,
            tpr_recall,
            step='post',
            alpha=0.2,
            color=roc_color)
        
        # diagonal line
        ax[0].plot(
            [0, 1], 
            [0, 1], 
            color=neutral_color,
            lw=lw,
            alpha=0.6,
            linestyle=neutral_linestyle)
        
        ax[0].set_xlim([-0.1, 1.1])
        
        ax[0].set_ylim([-0.1, 1.1])
        
        ax[0].set_xlabel('Fallout', fontweight='bold')
        
        ax[0].set_ylabel('Recall', fontweight='bold')
        
        ax[0].set_title(
            'AUROC = {0:.2f}'.format(
                skm.auc(
                    x=fpr, 
                    y=tpr_recall)),
            fontweight='bold')

        #Plot dots at certain decision thresholds, for clarity
        for threshold in [.1, .3, .5, .7, .9]:
            
            tpr_recall, fpr, _ = BinaryClassifierHelper.calculate_tpr_fpr_prec(
                threshold=threshold,
                y_true=y_true,
                y_score=y_score)
            
            ax[0].plot(
                fpr, 
                tpr_recall, 
                'o', 
                color=roc_color)
            
            ax[0].annotate(
                '  t = {0:.1f}  '.format(threshold), 
                (fpr, tpr_recall),
                ha='right',
                rotation=-45,
                size=16)

            
        precision, tpr_recall, _ = skm.precision_recall_curve(
            y_true=y_true,
            probas_pred=y_score)
            
        ax[1].step(
            tpr_recall,
            precision,
            color=pr_color,
            alpha=0.2,
            where='post',
            linewidth=lw,
            linestyle=main_linestyle)
            
        ax[1].fill_between(
            tpr_recall,
            precision,
            step='post',
            alpha=0.2,
            color=pr_color)
            
        ax[1].set_xlabel('Recall', fontweight='bold')
            
        ax[1].set_ylabel('Precision', fontweight='bold')
            
        ax[1].set_xlim([-0.1, 1.1])
            
        ax[1].set_ylim([-0.1, 1.1])
            
        ax[1].set_title(
            'Average Precision = {0:.2f}'.format(
                skm.average_precision_score(
                    y_true=y_true,
                    y_score=y_score)), 
            fontweight='bold')

        #Plot dots at certain decision thresholds, for clarity
        for threshold in [.1, .3, .5, .7, .9]:
            
            tpr_recall, _, precision = BinaryClassifierHelper.calculate_tpr_fpr_prec(
                threshold=threshold,
                y_true=y_true,
                y_score=y_score)
            
            ax[1].plot(
                tpr_recall, 
                precision, 
                'o',
                color=pr_color)
            
            ax[1].annotate(
                '  t = {0:.1f}  '.format(threshold), 
                (tpr_recall, precision),
                ha='left',
                rotation=45,
                size=16)

        # Recall / Precision: xing curves
        threshold_value = list(np.linspace(
            start=0.0, 
            stop=1.0, 
            num=200))
        
        tpr_v = []
        ppv_v = []
        
        for threshold in threshold_value:
            
            tpr, _, ppv = BinaryClassifierHelper.calculate_tpr_fpr_prec(
                threshold=threshold,
                y_true=y_true,
                y_score=y_score)
            
            tpr_v.append(tpr)
            ppv_v.append(ppv)
                    
        ax[2].plot(
            threshold_value,
            tpr_v,
            label='Recall')

        ax[2].plot(
            threshold_value, 
            ppv_v, 
            label='Precision')

        if best_threshold is not None:
            
            x_opt = best_threshold

            tpr_opt, _, ppv_opt = BinaryClassifierHelper.calculate_tpr_fpr_prec(
                threshold=best_threshold,
                y_true=y_true,
                y_score=y_score)
            
            ax[2].plot(
                x_opt, 
                tpr_opt, 
                'o',
                color=pr_color)

            ax[2].plot(
                x_opt, 
                ppv_opt, 
                'o',
                color=pr_color)

            ax[2].plot(
                x_opt, 
                0, 
                'o',
                color=pr_color)
            
            ax[2].plot(
                [x_opt, x_opt], 
                [0, max(tpr_opt, ppv_opt)], 
                color=neutral_color,
                lw=lw,
                alpha=0.6,
                linestyle=neutral_linestyle)

            ax[2].annotate(
                '  t = {0:.4f}  '.format(x_opt), 
                (x_opt, 0),
                ha='left',
                rotation=45,
                size=16)
            
        ax[2].set_xlabel('Threshold', fontweight='bold')
        ax[2].set_ylabel('Metric', fontweight='bold')
        ax[2].set_xlim([-0.1, 1.1])
        ax[2].set_ylim([-0.1, 1.1])
        ax[2].legend(loc='upper center')
        
        ax[2].set_title(
            'Threshold Maximizing F1-Score = {0:.4f}'.format(x_opt), 
            fontweight='bold')

        fraction_of_positives, mean_predicted_value = skc.calibration_curve(
            y_true=y_true,
            y_prob=y_score,
            n_bins=n_bins)
        
        # Reliability Curve (Calibration)
        ax[3].plot(
            [0, 1],
            [0, 1],
            'k:', 
            label='Perfectly calibrated')
        
        ax[3].plot(
            mean_predicted_value, 
            fraction_of_positives,
            'o-',
            color=gph.GraphicsStatics.get_color(1),
            label='Current')
        
        ax[3].set_xlabel('Mean predicted value', fontweight='bold')
        ax[3].set_ylabel('Fraction of positives', fontweight='bold')
        ax[3].set_ylim([-0.05, 1.05])
        ax[3].legend(loc='lower right')
        
        ax[3].set_title(
            'Reliability Curve (Calibration)', 
            fontweight='bold')

        
        # Probability Distribution
        ax[4].hist(
            y_score,
            range=(0, 1),
            log=True,
            bins=n_bins,
            histtype='bar',
            lw=2)    

        ax[4].set_xlabel('Mean predicted value', fontweight='bold')
        ax[4].set_ylabel('Count (log)', fontweight='bold')
        
        ax[4].set_title(
            'Probability Distribution', 
            fontweight='bold')
        
        
        plt.subplots_adjust(
            top=0.91,
            wspace=.2, 
            hspace=.2)
        
        fig.suptitle(
            t=title, 
            fontsize=26, 
            fontweight='bold',
            verticalalignment='top',
            horizontalalignment='center',
            fontstyle='italic',
            x=(fig.subplotpars.right + fig.subplotpars.left)/2)

        fig.text(
            x=0.02, 
            y=1, 
            s=BinaryClassifierHelper.get_confusion_matrix_string(
                threshold=best_threshold,
                y_true=y_true,
                y_score=y_score), 
            ha='left', 
            va='top',
            fontsize=14,
            fontstyle='italic',
            bbox={ 
                'facecolor': gph.GraphicsStatics.get_color(5), 
                'alpha': 0.5, 
                'pad': 7
            })

        fig.text(
            x=.95, 
            y=1, 
            s=BinaryClassifierHelper.get_metrics_string(
                threshold=best_threshold,
                y_true=y_true,
                y_score=y_score), 
            ha='right', 
            va='top',
            fontsize=14,
            fontstyle='italic',
            bbox={ 
                'facecolor': gph.GraphicsStatics.get_color(6), 
                'alpha': 0.5, 
                'pad': 7
            })    
            
        plt.show()



import logging
import os as os
import tempfile
import datetime as dt
from matplotlib import pyplot as plt

import neptune as npt

from neptunecontrib import api as npta
from neptunecontrib import hpo as npth
from neptunecontrib.monitoring import optuna as nmo
from neptunecontrib.monitoring import keras as nmk


class ExperimentManager(object):

    def __init__(self, log, project_name, experiment_name, experiment_params, experiment_tags):

        logging.getLogger('neptune').setLevel(logging.ERROR)

        self.log = log
        self.project = None
        self.experiment = None
        self._optuna_callback = None
        self._keras_callback = None
        self._experiment_name = None

        if self.log:
            self.project=npt.set_project(
                project_qualified_name=project_name)

            self.experiment = npt.create_experiment(
                name=experiment_name,
                params=experiment_params)

            for tag in experiment_tags:
                self.experiment.append_tags(tag)

    @property
    def experiment_name(self):
        if self._experiment_name is None:
            if self.log and self.experiment:
                self._experiment_name = self.experiment.id
            else:
                self._experiment_name = ExperimentManager.get_fake_name()

        return self._experiment_name

    @property
    def optuna_callback(self):
        if self.log and self._optuna_callback is None:
                self._optuna_callback = nmo.NeptuneCallback(
                    experiment=self.experiment)

        return self._optuna_callback

    @property
    def keras_callback(self):
        if self.log and self._keras_callback is None:
                self._keras_callback = nmk.NeptuneCallback(
                    experiment=self.experiment)

        return self._keras_callback

    @staticmethod
    def get_fake_name():
        return str(int(dt.datetime.now().timestamp()))

    def get_experiment_id(self):
        if self.log and self.experiment:
            return self.experiment.id
        else:
            return None

    def add_experiment_tags(self, tag):
        if self.log and self.experiment:
            self.experiment.append_tags(tag)

            print('[Tag] {0}'.format(tag))

    def set_experiment_property(self, key, value):
        if self.log and self.experiment:
            self.experiment.set_property(
                key=key, 
                value=value)

            print('[Property] {0}: {1}'.format(key, value))

    def log_data_to_neptune(self, data, name, log_index=False, log_to_table=True, log_to_artifact=True):
        if self.log and self.experiment:
            if name is None:
                name = ExperimentManager.get_fake_name()

            if log_to_table:
                
                npta.log_table(
                    name='{0}.csv'.format(name), 
                    table=data,
                    experiment=self.experiment)

            if log_to_artifact:
                
                file_path = os.path.join(
                    tempfile.gettempdir(),
                    '{0}.csv'.format(name))

                data.to_csv(file_path, index=log_index)

                self.experiment.log_artifact(
                    file_path,
                    'data/{0}.csv'.format(name))

    def log_chart_to_neptune(self, figure, name, log_to_artifact=True):
        if self.log and self.experiment:
            if name is None:
                name = ExperimentManager.get_fake_name()

            if log_to_artifact:
                file_path = os.path.join(
                    tempfile.gettempdir(),
                    '{0}.jpg'.format(name))

                figure.savefig(
                    fname=file_path, 
                    dpi=96,
                    format='jpg',
                    quality=90,
                    bbox_inches='tight')

                self.experiment.log_artifact(
                    file_path,
                    'charts/{0}.jpg'.format(name))



import math as m
import pandas as pd
import numpy as np

from sklearn import preprocessing as skp
from sklearn import pipeline as skpl

from ..utilities import commonhelpers as cmn


def add_distributed_noise(ts, noise_std=0.05, noise_mean=0.0, noise_ratio=1.05, rng=None, random_seed=None):
    """Adds white noise to a time series 1D array.

    Args:
        ts (1D numpy array): time series data.
        noise_std (float, optional): standard deviation of the added noise. Defaults to 0.05.
        noise_mean (float, optional): mean of the added noise. Defaults to 0.0.
        noise_ratio (float, optional): ratio of the overall added noise. Defaults to 1.05.
        rng ([type], optional): random number generator. Will be created if None is passed. Defaults to None.
        random_seed ([type], optional): random seed to be used by the random generator. Defaults to None.

    Returns:
        (1D numpy array): same shape as `ts`. Time series with white noise added to it.
    """
    ts_mean = ts.mean()
    ts_std = ts.std()
    
    ts_scaled = (ts - ts_mean) / ts_std
    ts_length = len(ts_scaled)
    
    if rng is None:
        rng = np.random.default_rng(
            seed=random_seed)
    
    ts_scaled_noised = noise_ratio * (noise_std * rng.standard_normal(ts_length) + noise_mean) + ts_scaled
    
    return ts_std * ts_scaled_noised + ts_mean

def prepare_data(data, lags_in, cols_in, steps_in, cols_out, steps_out, augmentation_factor=0, noise_std=0.05, noise_mean=0.0, noise_ratio=1.05, verbose=1):
    """[summary]

    Args:
        data (DataFrame): [description]
        lags_in (list): lagged versions of input data to inject as input. If `None`, will be set to `[0]`. If a `list` is passed, value `0` is prepended to the list (and named `lags`).
        cols_in (list): list of columns to use as input.
        steps_in (int): number of input steps.
        cols_out (list): list of columns to use as output.
        steps_out (int): number of output steps.
        augmentation_factor (int, optional): Number of time the original dataset is replicated with noise addition. Defaults to 0.
        noise_std (float, optional): standard deviation of the added noise. Defaults to 0.05.
        noise_mean (float, optional): mean of the added noise. Defaults to 0.0.
        noise_ratio (float, optional): ratio of the overall added noise. Defaults to 1.05.
        verbose (int, optional): [description]. Defaults to 1.

    Raises:
        ValueError: in case there isn't enough data to produce a single sample.

    Returns:
        X (3D array of shape (samples * (augmentation_factor + 1), steps_in, len(cols_in) * len(lags))):
        y (3D array of shape (samples * (augmentation_factor + 1), steps_out, len(cols_out))) 
        index (list): the input data index.
        scaler_in (scikit-learn transformer): transformer used to fit and transform input data X.
        scaler_out (scikit-learn transformer): transformer used to fit and transform output data y.
        samples (int): the number of produced samples before data augmentation.
    """    

    lags = [0]

    if lags_in is not None:
        # Contains all lag values (even the value 0: no-lag)    
        lags = list(set(lags + lags_in))

    # Maximum lag value
    steps_lag = max(lags)

    # Input data feeds
    data_in = data[cols_in].values

    if verbose > 0:
        print('data_in.shape = {0}'.format(data_in.shape))
        print('data_in contains np.nan = {0}, count() = {1}'.format(
            np.isnan(data_in).max(),
            np.isnan(data_in).sum()))

    # Output data feeds
    data_out = data[cols_out].values

    if verbose > 0:
        print('data_out.shape = {0}'.format(data_out.shape))
        print('data_out contains np.nan = {0}, count() = {1}'.format(
            np.isnan(data_out).max(),
            np.isnan(data_out).sum()))

    # This is be the number of samples that can be built out of the input dataframe
    samples = data_in.shape[0] - (steps_in + steps_out + steps_lag) + 1

    if verbose > 0:
        print('samples = {0}'.format(samples))
    
    if samples < 1:
        raise ValueError('not enough data to produce 1 sample.')
    
    # Information to retain for return of function
    index = list(data.index)
    
    # Scaler for input data
    scaler_in = skpl.Pipeline([
        ('std', skp.StandardScaler()),
        ('minmax', skp.MinMaxScaler(feature_range=(0, 1)))])
    
    data_in_scaled = scaler_in.fit_transform(data_in)

    if verbose > 0:
        print('data_in_scaled.shape = {0}'.format(data_in_scaled.shape))
        print('data_in_scaled contains np.nan = {0}, count() = {1}'.format(
            np.isnan(data_in_scaled).max(),
            np.isnan(data_in_scaled).sum()))
    
    # Scaler for output data
    scaler_out = skpl.Pipeline([
        ('std', skp.StandardScaler()),
        ('minmax', skp.MinMaxScaler(feature_range=(0, 1)))])

    data_out_scaled = scaler_out.fit_transform(data_out)

    if verbose > 0:
        print('data_out_scaled.shape = {0}'.format(data_out_scaled.shape))
        print('data_out_scaled contains np.nan = {0}, count() = {1}'.format(
            np.isnan(data_out_scaled).max(),
            np.isnan(data_out_scaled).sum()))

    # Generated dataset   
    X = []
    y = []
    
    for sample in range(samples):

        for step_in in range(steps_in):

            for index_in in range(len(cols_in)):

                index_data_in_scaled = data_in_scaled[:, index_in]
                
                for lag in lags:

                    X.append(
                        index_data_in_scaled[sample + steps_lag + step_in - lag])

        for step_out in range(steps_out):

            for index_out in range(len(cols_out)):

                index_data_out_scaled =  data_out_scaled[:, index_out]

                y.append(
                    index_data_out_scaled[sample + steps_lag + steps_in + step_out])
    
    # Check if data augmentation is required
    if augmentation_factor is None:
        augmentation_factor = 0

    if augmentation_factor > 0:

        for _ in range(augmentation_factor):

            for sample in range(samples):

                for step_in in range(steps_in):

                    for index_in in range(len(cols_in)):

                        index_data_in_scaled = add_distributed_noise(
                            ts=data_in_scaled[:, index_in],
                            noise_std=noise_std,
                            noise_mean=noise_mean,
                            noise_ratio=noise_ratio)
                    
                        for lag in lags:

                            X.append(
                                index_data_in_scaled[sample + steps_lag + step_in - lag])

                for step_out in range(steps_out):

                    for index_out in range(len(cols_out)):

                        index_data_out_scaled =  add_distributed_noise(
                            ts=data_out_scaled[:, index_out],
                            noise_std=noise_std,
                            noise_mean=noise_mean,
                            noise_ratio=noise_ratio)

                        y.append(
                            index_data_out_scaled[sample + steps_lag + steps_in + step_out])

    X = np.array(X).reshape(
        samples * (augmentation_factor + 1), 
        steps_in, 
        len(cols_in) * len(lags))

    if verbose > 0:
        print('X.shape = {0}'.format(X.shape))
        print('X contains np.nan = {0}, count() = {1}'.format(
            np.isnan(X).max(),
            np.isnan(X).sum()))

    y = np.array(y).reshape(
        samples * (augmentation_factor + 1), 
        steps_out,
        len(cols_out))

    if verbose > 0:
        print('y.shape = {0}'.format(y.shape))
        print('y contains np.nan = {0}, count() = {1}'.format(
            np.isnan(y).max(),
            np.isnan(y).sum()))
    
    return X, y, index, scaler_in, scaler_out, samples


from __future__ import absolute_import

from . import classificationhelpers
from . import commonhelpers
from . import filehelpers
from . import graphichelpers
from . import httphelpers
from . import imagehelpers
from . import mssqlhelpers
from . import neptunehelpers



from __future__ import absolute_import

from .classifiers import embedders
from .classifiers import features
from .classifiers import hierarchical

from .timeseries import dataaugmentationhelpers
from .timeseries import prophethelpers
from .timeseries import visualizationhelpers

from .utilities import classificationhelpers
from .utilities import commonhelpers
from .utilities import filehelpers
from .utilities import graphichelpers
from .utilities import httphelpers
from .utilities import imagehelpers
from .utilities import mssqlhelpers
from .utilities import neptunehelpers

__version__ = '0.9.9'


import pandas as pd
import numpy as np
import fbprophet as fbp
import matplotlib as mat
from matplotlib import pyplot as plt

from ..utilities import graphichelpers as gph


def regressor_index(m, name):
    """Given the name of a regressor, return its (column) index in the `beta` matrix.
    Parameters
    ----------
    m: Prophet model object, after fitting.
    name: Name of the regressor, as passed into the `add_regressor` function.
    Returns
    -------
    The column index of the regressor in the `beta` matrix.
    """
    return np.extract(
        m.train_component_cols[name] == 1, m.train_component_cols.index
    )[0]

def regressor_coefficients(m):
    """Summarise the coefficients of the extra regressors used in the model.
    For additive regressors, the coefficient represents the incremental impact
    on `y` of a unit increase in the regressor. For multiplicative regressors,
    the incremental impact is equal to `trend(t)` multiplied by the coefficient.
    Coefficients are measured on the original scale of the training data.
    Parameters
    ----------
    m: Prophet model object, after fitting.
    Returns
    -------
    pd.DataFrame containing:
    - `regressor`: Name of the regressor
    - `regressor_mode`: Whether the regressor has an additive or multiplicative
        effect on `y`.
    - `center`: The mean of the regressor if it was standardized. Otherwise 0.
    - `coef_lower`: Lower bound for the coefficient, estimated from the MCMC samples.
        Only different to `coef` if `mcmc_samples > 0`.
    - `coef`: Expected value of the coefficient.
    - `coef_upper`: Upper bound for the coefficient, estimated from MCMC samples.
        Only to different to `coef` if `mcmc_samples > 0`.
    """
    assert len(m.extra_regressors) > 0, 'No extra regressors found.'
    
    # trend_mean = m.params['trend'].mean()
    trend_std = m.params['trend'].std()
    
    coefs = []
    
    for regressor, params in m.extra_regressors.items():
        
        beta = m.params['beta'][:, regressor_index(m, regressor)]
        additive = (params['mode'] == 'additive')
        
        if params['mode'] == 'additive':
            coef = beta * m.y_scale / params['std']
        else:
            coef = beta / params['std']
            
        percentiles = [
            (1 - m.interval_width) / 2,
            1 - (1 - m.interval_width) / 2,
        ]
        
        coef_bounds = np.quantile(coef, q=percentiles)
        
        record = {
            'regressor': regressor,
            'regressor_mode': params['mode'],
            'center': float(params['mu']),
            'coef_lower': coef_bounds[0],
            'coef': np.mean(coef),
            'coef_upper': coef_bounds[1],
            'beta': float(beta[0]),
            'beta_abs': abs(float(beta[0])),
            'beta_lower': float(beta[0]) if additive else float(beta[0]) - trend_std,
            'beta_upper': float(beta[0]) if additive else float(beta[0]) + trend_std
        }
        
        coefs.append(record)

    return pd.DataFrame(coefs)

def plot_regressors_importance(m, title=None, subtitle=None, name=None, regressors_names=None, y_name='beta_abs', y_label='Importance %', experiment_manager=None, display=True):
    
    df = regressor_coefficients(m).sort_values(by=y_name, ascending=False)
    df = df[['regressor', y_name]]
    df = df.rename(columns={ y_name: 'importance' })
    df.loc[:, 'importance'] = df['importance']

    # syle
    gph.GraphicsStatics.initialize_matplotlib_styles()

    # Create the figure and the axes
    fig, ax = plt.subplots(
        figsize=gph.GraphicsStatics.g_landscape_fig_size)
    
    x = range(df.shape[0])
    y = list(np.array(df['importance']))
    
    rects = ax.bar(
        x=x,
        height=y)

    ax.set_xticks(x)

    labels = list(df['regressor'])
    
    if regressors_names is not None:
        labels = [regressors_names[label] for label in labels]
        
    ax.set_xticklabels(
        labels=labels)  

    ax.set(
        ylabel=y_label)

    for i, rect in enumerate(rects):
        
        height = rect.get_height()
        
        ax.annotate(
            '{0:.2%}'.format(y[i]),
            xy=(rect.get_x() + rect.get_width() / 2, height),
            xytext=(0, 3),
            textcoords="offset points",
            ha='center',
            va='bottom')

    gph.GraphicsStatics.style_plot(
        title=title, 
        subtitle=subtitle, 
        tl_name=name,
        fig=fig,
        ax=ax,
        experiment_manager=experiment_manager)

    if display:
        plt.show()
    else:
        return fig
